#!/usr/bin/env python3

# The command-line runner for the ffmirror program. Since the module and mirror
# APIs are by necessity somewhat primitive, this contains several
# user-interface functions that make the semantics nice and easy. These are
# designed for command-line use; other applications should probably use the
# separate module APIs.

import sys, argparse, os, json
from . import util, mirror, metadb, site_modules
from .core import url_res, DownloadModule
from typing import Optional
import click

cur_mirror = mirror.FFMirror('.', use_ids=True)

def parse_url(url: str) -> Optional[DownloadModule]:
    """Given a fanfiction URL (either story or author), this function will return
    the module that deals with the correct site. It does so by checking the
    regexes in urlres.

    """
    for reg, mod in url_res:
        if reg.match(url):
            return mod
    return None

def download_story(url: str, silent: bool = False,
                   contents: bool = True, update: bool = False,
                   dry_run: bool = False,
                   outfile: Optional[str] = None) -> None:
    if update:
        t = mirror.read_from_file(url)
        if t is None:
            print(f"Couldn't read file {url}", file=sys.stderr)
            sys.exit(1)
        ufn = url
        mod = site_modules[t.site]
        url = mod.get_story_url(t)
    else:
        tv = parse_url(url)
        assert tv is not None
        mod = tv
    o = mod.story_url_re.match(url)
    assert o is not None
    sid = o.group('sid')
    md, toc = mod.download_metadata(sid)
    if not silent:
        print("Found story {}, {} chapters".
              format(md.title, md.chapters))
    if dry_run:
        print('\nMetadata:')
        d = md.to_dict()
        for i in d:
            print("{}: {}".format(i, d[i]))
        print("\nContents:")
        for c in toc:
            print(c)
        return
    if update:
        if not cur_mirror.check_update(md, ufn):
            if not silent:
                print("Nothing to do (up to date)")
            return
    if outfile:
        fn = outfile
    else:
        fn = util.make_filename(md.title) + ".html"
    with open(fn, 'w') as out_fo:
        def progress(n, t):
            if not silent:
                print("Got chapter {} of {}".format(n + 1, len(toc)), end='\r')
        mod.compile_story(md, toc, out_fo, callback=progress,
                          contents=contents)
        if not silent:
            print("", end='\n')

def run_dl() -> None:
    ap = argparse.ArgumentParser(
        description="Download a single story as a plain file")
    ap.add_argument("-s", "--silent", action="store_true",
                    help="Suppress running output", default=False)
    g = ap.add_mutually_exclusive_group()
    ap.add_argument("-o", "--outfile", help="The file to output to",
                    default="")
    g.add_argument("-c", "--contents", action="store_true",
                   help="Generate a table of contents", default=False)
    ap.add_argument("-d", "--dry-run", action="store_true",
                    help="Dry run (no download, just parse metadata)",
                    default=False)
    ap.add_argument("-u", "--update", action="store_true",
                    help="Update an existing file", default=False)
    ap.add_argument("url", help="A URL for a chapter of the fic, or (with -u) filename for update",  # noqa: E501
                    default=None)
    args = ap.parse_args()
    download_story(**args.__dict__)

def download_list(url, ls=False, silent=False, getall=False, dry_run=False,
                  write_favs=False, **kwargs) -> None:
    mod = parse_url(url)
    assert mod is not None
    o = mod.user_url_re.match(url)
    assert o is not None
    uid = o.group('aid')
    auth, fav, info = mod.download_list(uid)
    if write_favs:
        dn = "{}-{}-{}".format(util.make_filename(info.name),
                               info.site, info.id)
        os.makedirs(dn, exist_ok=True)
        with open(os.path.join(dn, 'favorites.json'), 'w') as out:
            json.dump({'info': info, 'favs': fav}, out, sort_keys=True,
                      indent=1)
    sl = fav if ls else auth
    if not getall:
        nsl = [i for i in sl if cur_mirror.check_update(i)]
    else:
        nsl = sl
    if not silent and len(auth) > 0:
        print("Got {} (of {}) stories from author {}".
              format(len(nsl), len(sl), info.name))
    if dry_run:
        for i in sl:
            print(i)
        return
    cur_mirror.update_tags(nsl)
    lsl = 0

    def progress(i, n) -> None:
        nonlocal lsl
        if not silent:
            if type(n) == tuple:
                print("Acquiring story '{}' (#{}/{})".
                      format(n[0]['title'], i + 1, len(nsl)))
                lsl = len(n[1])
            else:
                print("Got chapter {} of {}".format(i + 1, lsl), end='\r')
                if i == lsl - 1:
                    print("\n", end="")
    cur_mirror.update_list(nsl, callback=progress)

def run_add() -> None:
    ap = argparse.ArgumentParser(description="Add an author's corpus or favorites to a mirror, or update them")  # noqa: E501
    ap.add_argument("-s", "--silent", action="store_true",
                    help="Suppress running output", default=False)
    ap.add_argument("-f", "--favorites", dest='ls', action="store_true",
                    default=False,
                    help="Get author's favorites rather than their corpus")
    ap.add_argument("-a", "--all", dest='getall', action="store_true",
                    default=False,
                    help="Download all stories without checking if already present")  # noqa: E501
    ap.add_argument("-d", "--dry-run", action="store_true", default=False,
                    help="Dry run (only parse list and print)")
    ap.add_argument("url", help="A URL for an author's profile")
    args = ap.parse_args()
    download_list(write_favs=True, **args.__dict__)

def update_mirror(silent=False, author=None) -> None:
    m = cur_mirror.read_entries()
    for n, i in enumerate(sorted(m.keys())):
        if author is not None and author != i:
            continue
        if not silent:
            print("Author '{}' (#{}/{})".format(m[i].info['author'], n + 1,
                                                len(m)))
        url = m[i].info['author_url']
        download_list(url, write_favs=True, silent=silent)

def run_update() -> None:
    ap = argparse.ArgumentParser(description="Update an entire mirror from the Web site")  # noqa: E501
    ap.add_argument("-s", "--silent", action="store_true",
                    help="Suppress running output", default=False)
    ap.add_argument("author", nargs='?', help="Update only a given author",
                    default=None)
    args = ap.parse_args()
    update_mirror(**args.__dict__)

def do_cache() -> None:
    cur_mirror.make_cache()

@click.group()
def run_db_op() -> None:
    pass

@run_db_op.command()
@click.option("-a", "--author-dir", type=str,
              help="Author directory to update", default=None)
@click.option("-m", "--max-authors", type=int,
              help="Maximum number of authors to update "
              "(to dodge Cloudflare rate-limiting)",
              default=None)
def update(author_dir: Optional[str], max_authors: Optional[int]) -> None:
    """Update the DB. If author-dir is given, update only that author."""
    mm = metadb.DBMirror('.')
    mm.connect()
    if author_dir is not None:
        dn = author_dir
        if dn.endswith('/'):
            dn = dn[:-1]
        site, aid = metadb.get_archive_id(dn)
        ao = mm.get_author(site, aid)
        mm.sync_author(ao)
        mm.archive_author(ao)
    else:
        mm.run_update(max_authors=max_authors)

@run_db_op.command()
@click.argument("url", type=str)
def add(url: str):
    """Add a new author to the DB."""
    mm = metadb.DBMirror('.')
    mm.connect()
    mod = parse_url(url)
    assert mod is not None
    site = mod.this_site
    o = mod.user_url_re.match(url)
    assert o is not None
    aid = o.group('aid')
    mm.sync_author((site, aid))
    ao = mm.get_author(site, aid)
    mm.archive_author(ao)

@run_db_op.command()
def init():
    """Initialize an ffmirror database in the current directory."""
    mm = metadb.DBMirror('.')
    mm.connect()
    mm.create()



# Functions for maintaining a metadata-only mirror in an SQLAlchemy database.
# This may be more feasible to duplicate in its entirety offline.

from __future__ import annotations

from . import site_modules
from .core import StoryInfo, AuthorInfo

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import (sessionmaker, relationship,  # noqa: F401
                            joinedload, exc)
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy import create_engine, text, func  # noqa: F401
from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy import Table, DateTime, Boolean, Interval
from sqlalchemy.engine.base import Engine

from typing import Union, Tuple, Optional, List, cast, Set

import datetime, os, traceback, sys

db_file = 'db_test.sqlite'

Base = declarative_base()

fav_stories_table = Table('fav_stories', Base.metadata,
                          Column('author_id', Integer,
                                 ForeignKey('author.id')),
                          Column('story_id', Integer, ForeignKey('story.id')))

fav_authors_table = Table('fav_authors', Base.metadata,
                          Column('author_id', Integer,
                                 ForeignKey('author.id')),
                          Column('favauthor_id', Integer,
                                 ForeignKey('author.id')))

story_tags_table = Table('story_tags', Base.metadata,
                         Column('story_id', Integer, ForeignKey('story.id')),
                         Column('tag_id', Integer, ForeignKey('tag.id')))

class Author(Base):
    __tablename__ = 'author'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    archive = Column(String, nullable=False)
    site_id = Column(String, nullable=False)

    md_synced = Column(DateTime)
    sync_int = Column(Interval)

    in_mirror = Column(Boolean, nullable=False, default=False)

    stories_written: RelationshipProperty[List[Story]] = relationship(
        'Story', back_populates='author'
    )
    fav_stories: RelationshipProperty[List[Story]] = relationship(
        'Story', secondary=fav_stories_table,
        backref='users_faved'
    )
    fav_authors: RelationshipProperty[List[Author]] = relationship(
        'Author', secondary=fav_authors_table,
        primaryjoin=id == fav_authors_table.c.author_id,
        secondaryjoin=id == fav_authors_table.c.favauthor_id,
        backref='users_faved'
    )

    def __repr__(self):
        return "<Author name='{}' id='{}'>".format(self.name, self.site_id)

    def get_metadata(self) -> AuthorInfo:
        mod = site_modules[self.archive]
        # md = { 'authorid': self.site_id }
        authinf = AuthorInfo(name=self.name, id=self.site_id,
                             url='', site=self.archive, dir='')
        authinf.dir = authinf.get_mirror_dirname()
        authinf.url = mod.get_user_url(authinf)
        return authinf

    def source_site_url(self):
        return self.get_metadata().url

class Story(Base):
    __tablename__ = 'story'

    id = Column(Integer, primary_key=True)
    title = Column(String)
    archive = Column(String, nullable=False)
    site_id = Column(String, nullable=False)
    words = Column(Integer)
    chapters = Column(Integer)
    published = Column(DateTime)
    updated = Column(DateTime)
    category = Column(String)
    summary = Column(String)
    characters = Column(String)
    complete = Column(Boolean)
    genre = Column(String)

    download_time = Column(DateTime)  # null if never downloaded
    download_fn = Column(String)

    author_id = Column(Integer, ForeignKey('author.id'))

    author = relationship('Author', back_populates='stories_written')

    tags = relationship('Tag', secondary=story_tags_table,
                        backref='stories', uselist=True)

    def get_metadata(self) -> StoryInfo:
        site = self.archive
        mod = site_modules[site]
        authinf = self.author.get_metadata()
        chapters = self.chapters or -1
        words = self.words or -1
        assert self.updated is not None
        assert self.published is not None
        ts = cast(Set[str], set(i.name for i in self.tags if i))
        rv = StoryInfo(
            title=self.title, summary=self.summary, category=self.category,
            id=self.site_id, reviews=-1, chapters=chapters, words=words,
            characters=self.characters, source='', author=authinf,
            genre=self.genre, site=self.archive, updated=self.updated,
            published=self.published, complete=self.complete, story_url='',
            tags=ts
        )
        rv.story_url = mod.get_story_url(rv)
        return rv

    def unique_filename(self) -> str:
        tmd = self.get_metadata()
        # return story_file(tmd, with_id=True)
        return tmd.get_mirror_filename()

    def __repr__(self) -> str:
        return "<Story '{}' by '{}' id='{}'>".format(
            self.title, self.author.name, self.site_id)

class Tag(Base):
    __tablename__ = 'tag'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    date_added = Column(DateTime, default=datetime.datetime.now)

    def __repr__(self) -> str:
        return "<Tag '{}'>".format(self.name)

class Config(Base):
    __tablename__ = 'config'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    value = Column(String)

    def __repr__(self) -> str:
        return "<Config '{}'='{}'>".format(self.name, self.value)

class DBMirror(object):
    def __init__(self, mdir: str, debug: bool = False) -> None:
        self.engine: Optional[Engine] = None
        self.Session: Optional[sessionmaker] = None
        self.mdir = os.path.abspath(mdir)
        self.db_file = os.path.join(self.mdir, 'ffmeta.sqlite')
        self.debug = debug
        self._last_ao = None

    def __enter__(self) -> DBMirror:
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        # any method which changes the DB will call ds.commit() on its own
        self.ds.close()

    def connect(self) -> None:
        self.engine = create_engine('sqlite:///{}'.format(self.db_file),
                                    echo=self.debug)
        self.Session = sessionmaker(bind=self.engine)
        self.ds = self.Session()

    def create(self) -> None:
        Base.metadata.create_all(self.engine)
        co = Config(name='archive_dir', value=self.mdir)
        self.ds.add(co)
        self.ds.commit()

    def get_story(self, archive: str, sid: str) -> Story:
        so_rv = (self.ds.query(Story)
                 .filter((Story.archive == archive) & (Story.site_id == sid)).
                 one_or_none())
        return so_rv

    def get_author(self, archive: str, aid: str) -> Author:
        return (self.ds.query(Author)
                .filter_by(archive=archive).filter_by(site_id=aid).
                one_or_none())

    def cache_load_author(self, ao: Author) -> Author:
        return (self.ds.query(Author)
                .options(joinedload(Author.stories_written),
                         joinedload(Author.fav_stories).joinedload('author'))
                .filter_by(id=ao.id).one_or_none())

    def get_config(self, name: str) -> str:
        return self.ds.query(Config).filter_by(name=name).first().value

    def _handle_tags_for(self, md: StoryInfo) -> None:
        """Takes a story metadata object, fetches its tags via its respective site
        module, then adds them to that story in the database. Tags that do not
        exist will be created. Requires that the story object already exist.
        This method does not call ds.commit(); a caller MUST do so.

        """
        ds = self.ds
        so = self.get_story(md.site, md.id)
        for t in md.tags:
            to = ds.query(Tag).filter_by(name=t).first()
            if not to:
                to = Tag(name=t)
                ds.add(to)
            if so not in to.stories:
                to.stories.append(so)

    def _check_update(self, so: Story, s: StoryInfo) -> bool:
        return (so.words != s.words or so.chapters != s.chapters or
                so.updated != s.updated)

    def _story_from_md(self, s: StoryInfo, ao: Author,
                       eso: Story = None):
        """Gets or creates a Story object from an ffmirror metadata dictionary. If it
        did not already exist, the object is added to the database session. If
        it did, it is updated to match the parameters. This method does not
        call ds.commit(); a caller MUST do so.

        """
        ds = self.ds
        if eso:
            so = eso
        else:
            so = self.get_story(s.site, s.id)
        if not so:
            so = Story(archive=s.site, site_id=s.id)
            ds.add(so)
        if self._check_update(so, s):
            so.title = s.title
            so.words = s.words
            so.chapters = s.chapters
            so.category = s.category
            so.summary = s.summary
            so.characters = s.characters
            so.complete = s.complete
            so.genre = s.genre
            so.published = s.published
            so.updated = s.updated
            self._handle_tags_for(s)
            so.author = ao
        return so

    def sync_author(self, ido: Union[Author, Tuple[str, str]]) -> None:
        ds = self.ds
        if isinstance(ido, Author):
            ao = ido
            archive, aid = ao.archive, ao.site_id
        else:
            archive, aid = ido
            ao = self.get_author(archive, aid)
        mod = site_modules[archive]
        auth, fav, info = mod.download_list(aid)
        if not ao:
            ao = Author(name=info.name, archive=archive, site_id=aid)
            ao.sync_int = datetime.timedelta(days=1)
            ds.add(ao)
        else:
            ao = self.cache_load_author(ao)
        ao.md_synced = datetime.datetime.now()
        si_d = {}
        for s in ao.stories_written + ao.fav_stories:
            si_d[s.site_id] = s
        for sm in auth:
            so = self._story_from_md(sm, ao, si_d.get(sm.id))
        for sm in fav:
            ms = si_d.get(sm.id)
            fao = ms.author if ms else self.get_author(archive, sm.author.id)
            if not fao:
                fao = Author(name=sm.author.name, archive=archive,
                             site_id=sm.author.id)
                ds.add(fao)
            so = self._story_from_md(sm, fao, ms)
            if so not in ao.fav_stories:
                ao.fav_stories.append(so)
        ds.commit()

    def story_to_archive(self, i: Story, rfn: Optional[str] = None,
                         silent: bool = False, commit: bool = True) -> None:
        if not silent:
            print("Downloading story '{}'".format(i.title))
        mod = site_modules[i.archive]
        md, toc = mod.download_metadata(i.site_id)
        if rfn is None:
            rfn = md.get_mirror_filename()
        fn = os.path.join(self.mdir, rfn)
        os.makedirs(os.path.split(fn)[0], exist_ok=True)

        def ocf(n, t):
            if not silent:
                print("\r\x1b[2Kch.{}/{}: {}".format(n + 1,
                                                     i.chapters, t), end='')
        with open(fn, 'w') as out:
            mod.compile_story(md, toc, out, contents=True, callback=ocf)
        if not silent:
            print('', end='\n')
        i.download_fn = rfn
        i.download_time = datetime.datetime.now()
        if commit:
            self.ds.commit()

    def archive_author(self, ao: Author, silent: bool = False) -> None:
        ds = self.ds
        ao.in_mirror = True
        try:
            for i in ds.query(Story).filter((Story.author == ao) &
                                            ((Story.download_time == None) |  # noqa: E711,E501
                                             (Story.download_time <
                                              Story.updated))).all():
                try:
                    self.story_to_archive(i, silent=silent, commit=False)
                except Exception:
                    if not silent:
                        print("Download failed")
                        traceback.print_exc(file=sys.stdout)
        finally:
            ds.commit()

    def run_update(self, silent: bool = False,
                   max_authors: Optional[int] = None) -> None:
        ds = self.ds
        aq = (ds.query(Author).filter(Author.in_mirror == True).  # noqa: E712
              order_by(Author.md_synced.asc()))
        ta = aq.count()
        for n, i in enumerate(aq.all()):
            if max_authors is not None and n >= max_authors:
                break
            if not silent:
                print("Syncing author {} ({}/{})".format(i.name, n + 1, ta))
            self.sync_author(i)
            self.archive_author(i, silent=silent)

# Temp functions for mirror migrate.

def get_archive_id(dn: str) -> Tuple[str, str]:
    rv = dn.rsplit('-', 2)
    return rv[-2], rv[-1]

def populate_from_db(db, mm):
    import itertools
    all_stories = list(itertools.chain.from_iterable([i.stories for i in
                                                      db.values()]))
    for s in all_stories:
        so = mm.get_story(s['site'], s['id'])
        if not so:
            ao = mm.get_author(s['site'], s['authorid'])
            if not ao:
                raise ValueError("author not in db for story: {}".format(s))
            if 'complete' not in s:
                s['complete'] = False
            if type(s['complete']) != bool:
                s['complete'] = s['complete'] == 'True'
            if 'genre' not in s:
                s['genre'] = None
            so = mm._story_from_md(s, ao)
            print("Adding story not in db: {}".format(s))
        rfn = os.path.join(mm.mdir, s['filename'])
        if not os.path.exists(rfn):
            raise ValueError("file doesn't exist: {}".format(rfn))
        mt = datetime.datetime.fromtimestamp(os.stat(rfn).st_mtime)
        so.download_time = mt
        so.download_fn = s['filename']
    mm.ds.commit()



#!python3

from __future__ import annotations

from typing import (Dict, List, Tuple, Pattern, Any, TextIO, Callable, Set,
                    Optional)
from abc import ABCMeta, abstractmethod
from os.path import join
from .util import make_filename
import attr, datetime

@attr.s(auto_attribs=True)
class StoryInfo:
    title: str = attr.ib(converter=str)
    summary: str = attr.ib(converter=str)
    category: str = attr.ib(converter=str)
    id: str = attr.ib(converter=str)
    reviews: int = attr.ib(converter=int)
    chapters: int = attr.ib(converter=int)
    words: int = attr.ib(converter=int)
    characters: str = attr.ib(converter=str)
    source: str = attr.ib(converter=str)
    author: AuthorInfo = attr.ib()
    genre: str = attr.ib(converter=str)
    site: str = attr.ib(converter=str)
    updated: datetime.datetime
    published: datetime.datetime
    complete: bool = attr.ib(converter=bool)
    story_url: str = attr.ib(converter=str)
    tags: Set[str]

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> StoryInfo:
        authinf = AuthorInfo(
            name=d.pop('author'), id=d.pop('authorid'),
            url=d.pop('author_url'), site=d['site'],
            dir=d.pop('author_dir')
        )

        rv = cls(
            title=d.pop('title'), summary=d.pop('summary'),
            category=d.pop('category'), id=d.pop('id'),
            reviews=d.pop('reviews'), chapters=d.pop('chapters'),
            words=d.pop('words'), characters=d.pop('characters'),
            source=d.pop('source'), author=authinf, genre=d.pop('genre'),
            site=d.pop('site'), updated=d.pop('updated'),
            published=d.pop('published'), complete=d.pop('complete'),
            story_url=d.pop('story_url'), tags=d.pop('tags')
        )

        if len(d) > 0:
            raise ValueError("excess args in dict")

        return rv

    def to_dict(self) -> Dict[str, str]:
        adict = self.author.to_dict()

        rv = { k: str(v) for k, v in self.__dict__.items() }
        rv.pop('author')
        rv.update(adict)

        return rv

    def get_mirror_filename(self) -> str:
        return join(self.author.get_mirror_dirname(),
                    make_filename("{}-{}.html".format(self.title,
                                                      self.id)))

@attr.s(auto_attribs=True)
class ChapterInfo:
    title: str = attr.ib(converter=str)
    url: str = attr.ib(converter=str)

@attr.s(auto_attribs=True)
class AuthorInfo:
    name: str = attr.ib(converter=str)
    id: str = attr.ib(converter=str)
    url: str = attr.ib(converter=str)
    site: str = attr.ib(converter=str)
    dir: str = attr.ib(converter=str, default='')

    def to_dict(self) -> Dict[str, str]:
        return { 'author': self.name, 'authorid': self.id,
                 'author_url': self.url, 'author_dir': self.dir }

    def get_mirror_dirname(self) -> str:
        return "{}-{}-{}".format(make_filename(self.name),
                                 self.site, self.id)


site_modules: Dict[str, DownloadModule] = {}
url_res: List[Tuple[Pattern, DownloadModule]] = []

class TypeRegister(ABCMeta):
    def __init__(cls, name: str, bases: Tuple[type, ...],
                 datadict: Dict[str, Any]) -> None:
        super().__init__(name, bases, datadict)
        if 'this_site' in datadict and 'url_re' in datadict:
            nc = cls()
            site_modules[datadict['this_site']] = nc
            url_res.append((datadict['url_re'], nc))

class DownloadModule(metaclass=TypeRegister):
    story_url_re: Pattern
    user_url_re: Pattern
    this_site: str

    @abstractmethod
    def get_user_url(self, auth: AuthorInfo) -> str:
        ...

    @abstractmethod
    def get_story_url(self, story: StoryInfo) -> str:
        ...

    @abstractmethod
    def download_metadata(self, sid: str) -> Tuple[StoryInfo,
                                                   List[ChapterInfo]]:
        ...

    @abstractmethod
    def compile_story(self, story: StoryInfo, toc: List[ChapterInfo],
                      outfile: TextIO, contents: bool = True,
                      callback: Optional[Callable[[int, str], None]]
                      = None) -> None:
        ...

    @abstractmethod
    def download_chapter(self, chapter: ChapterInfo) -> str:
        ...

    @abstractmethod
    def download_list(self, aid: str) -> Tuple[List[StoryInfo],
                                               List[StoryInfo],
                                               AuthorInfo]:
        ...



#!/usr/bin/python

# This is a simple tagging system, originally part of the offline-dump
# maintainer of ffnet_scraper. Keeps tags in a text file, with a
# single entry per line, in the format '<resource>:
# <tag>[,<tag>,<tag>]'. Deals with tags objects which take the form of
# dictionaries from string resource to set of tags.

def read_tags(fn='tags'):
    """Reads a file containing tags, in the format 'resource:
    tag1,tag2,tag3' per line. Returns a dictionary of string to set of
    tags. If a resource shows up multiple times, just add all the tags
    in all appearances to one set. Tags may contain anything that
    isn't a comma, including whitespace; resources may contain
    anything that isn't a colon, but in practice are only ever
    alphanumeric plus [/_.]. The resource can be any string; in
    practice it's usually a filename for something."""
    if type(fn) == str:
        data = open(fn, "r")
    elif type(fn) == file:
        data = fn
    else:
        raise Exception("Invalid argument")
    d = {}
    for l in data:
        l = l[:-1] # clip newline
        i = l.find(": ")
        n, t = l.split(": ", 1)
        ts = [i for i in t.split(',') if len(i) > 0]
        if n in d:
            d[n].update(ts)
        else:
            d[n] = set(ts)
    return d

def get_tagset(tags):
    """Takes a tags object as returned by read_tags, returns a set of
    all tags used in the database."""
    s = set()
    for i in tags:
        s.update(tags[i])
    return s

def _extract_tobj(sobj):
    """Takes a list-of-stories dictionary as returned by read_entries,
    returns a tags object suitable for write_tags, by rebuilding the
    dictionary. This is a holdover from functionality in
    ffnet_scraper; it shouldn't be used in a tagging system."""
    rv = {}
    if type(sobj) == dict: # it's a straight RV dictionary
        for a, l in sobj.iteritems():
            for c in l:
                rv[c['filename']] = c['tags']
    elif type(sobj) == list: # it's a list of metadatas, more convenient to munge
        for c in sobj:
            rv[c['filename']] = c['tags']
    else:
        raise Exception("Invalid argument")
    return rv

def write_tags(tags, fn='tags'):
    """Writes a tags object (dictionary of sets, as the return value
    of read_tags() above) into the file tags."""
    if type(fn) == str:
        f = open(fn, "w")
    elif type(fn) == file:
        f = fn
    else:
        raise Exception("Invalid argument")
    for l in sorted(tags.keys()):
        f.write("{}: {}\n".format(l, ",".join(sorted(tags[l]))))
    if type(fn) == str:
        f.close()

def list_for_tag(tags, tag):
    """Takes a tags object as returned by read_tags and a tag name,
    returns a set of all the resources which have that tag."""
    rv = set()
    for f in tags.keys():
        if tag in tags[f]:
            rv.add(f)
    return rv



#!python3

from os.path import dirname, basename, isfile, join
import glob

modules = glob.glob(join(dirname(__file__), '*.py'))
__all__ = [ basename(f)[:-3] for f in modules if isfile(f) and not
            f.endswith('__init__.py') ]

from . import *



# Functions for maintaining an offline mirror of a large number of stories. The
# format of the mirror is as follows: in the top level are directories each
# titled for an author. Within each directory are files each titled for a
# story. Each file contains an entire story as HTML. The files contain comments
# with relevant metadata; this allows updating properly and browsing easily.
# The top-level directory also contains a single 'tags' file containing tags in
# the simpletags format; the resource is directory/file.html, while the tags
# are whatever.

import os, pickle, traceback, json
import ffmirror.util as util
from ffmirror.simpletags import read_tags, write_tags
from .core import StoryInfo
from typing import Optional, Dict, Any
from datetime import datetime
from . import site_modules as sites

def story_file(md: StoryInfo, with_id: bool = False) -> str:
    if with_id:
        return os.path.join(
            "{}-{}-{}".format(util.make_filename(md.author.name),
                              md.site, md.author.id),
            util.make_filename("{}-{}.html".format(md.title,
                                                   md.id)))
    else:
        return os.path.join(util.make_filename(md.author.name),
                            util.make_filename(md.title) + '.html')

def read_from_file(name: str) -> Optional[StoryInfo]:
    """Read a story metadata entry as returned by download_list out of
    a file. If file is nonexistent or unreadable, or metadata cannot
    be read, return None."""
    try:
        # Binary mode in order to support seek()
        f = open(name, "rb", buffering=0)
    except IOError:
        return None
    reading = False
    rv: Dict[str, Any] = {}
    for bline in f:
        line = bline.decode()
        # First (and presumably only) HTML comment in output file will be
        # metadata block
        if line[0:4] == "<!--" and "-->" not in line:
            reading = True
            continue
        if reading:
            try:
                k, v = line.strip("\n").split(': ', 1)
            except ValueError:
                break
            if k in ['published', 'updated']:
                try:
                    rv[k] = datetime.fromtimestamp(int(v))
                except ValueError:
                    rv[k] = datetime.fromisoformat(v)
            else:
                rv[k] = v
    if not reading:
        return None
    f.seek(-8, 2)
    s = f.readline().decode()
    # Incomplete file, probably left over from an earlier interrupted DL
    if s != "</html>\n":
        return None
    return StoryInfo.from_dict(rv)

# Turns out the optimized-and-kind-of-painful version really is about twice as
# fast as brute-force, which I think is worth it.

class AuthorInfo(object):
    def __init__(self, author_dir):
        self.stories = []
        self._info = None
        self._favs = None
        self._author_dir = author_dir

    @property
    def info(self):
        if self._info is not None:
            return self._info
        test = self.favs  # noqa: F841 (this is a hack)
        return self._info

    @property
    def favs(self):
        if self._favs is not None:
            return self._favs
        fn = os.path.join(self._author_dir, 'favorites.json')
        try:
            with open(fn) as inp:
                ai = json.load(inp)
        except:
            return []
        if self._info is None:
            self._info = ai['info']
        self._favs = ai['favs']
        return self._favs

    def with_tag(self, tag):
        rv = AuthorInfo(self._author_dir)
        rv.stories = [i for i in self.stories if tag in i['tags']]
        rv._info, rv._favs = self._info, self._favs
        if len(rv.stories) == 0:
            return None
        return rv

class FFMirror(object):
    def __init__(self, mirror_dir, use_ids=True):
        self.mirror_dir = mirror_dir
        self.use_ids = use_ids

    def check_update(self, r: StoryInfo, n: Optional[str] = None) -> bool:
        """Check a downloaded metadata entry r against local files. Return true if this
        entry needs redownloading. This will return false (thus failing) on
        colliding story titles if and only if use_ids is turned off. It is
        recommended to set use_ids to on.

        """
        if n is None:
            n = os.path.join(self.mirror_dir, story_file(r, self.use_ids))
        cr = read_from_file(n)
        if cr is None:
            return True
        # mod = sites[r.site]
        return not r == cr

    def update_list(self, sl, callback=None):
        """This function takes a list of stories (as metadata entries) and downloads
        them all. The filenames used are the result of story_file. No checking
        of update requirement is done; for update-only, the caller should
        filter on the result of check_update manually. callback is called at
        each story with the index and result of download_metadata for the
        current story; it is also passed to compile_story of the site module,
        which passes chapter index and string title.

        """
        for n, i in enumerate(sl):
            mod = sites[i['site']]
            try:
                md, toc = mod.download_metadata(i['id'])
            except Exception as e:
                print(i)
                traceback.print_exc()
                continue
            if callback:
                callback(n, (i, toc))
            fn = os.path.join(self.mirror_dir, story_file(md, self.use_ids))
            os.makedirs(os.path.split(fn)[0], exist_ok=True)
            with open(fn, 'w') as out:
                mod.compile_story(md, toc, out, contents=True,
                                  callback=callback)

    def update_tags(self, sl):
        """This function takes a list of metadata entries and updates the category tag
        on all of them. The result of cat_to_tagset on each story's category is
        added to that story's tag set; tags are then written back.

        """
        tfn = os.path.join(self.mirror_dir, 'tags')
        try:
            to = read_tags(tfn)
        except FileNotFoundError:
            to = {}
        for i in sl:
            fn = story_file(i, self.use_ids)
            mod = sites[i['site']]
            ct = mod.get_tags_for(i)
            if fn in to:
                to[fn].update(ct)
            else:
                to[fn] = ct
        write_tags(to, tfn)

    def read_entries(self):
        """Reads all the .html files below the current directory for ffmirror metadata;
        returns them as a dictionary of author name to list of story metadata
        entries. This should yield the data of all the files in the mirror.
        This function takes a long time on a large database; see the various
        caching functions.

        """
        rv = {}
        tfn = os.path.join(self.mirror_dir, 'tags')
        try:
            ts = read_tags(tfn)
        except FileNotFoundError:
            ts = {}
        for d, sds, fs in os.walk(self.mirror_dir):
            for i in [n for n in sds if n.startswith('.')]:
                sds.remove(i)
            for n in fs:
                if n.endswith(".html"):
                    fn = os.path.join(d, n)
                    rel_fn = os.path.relpath(fn, self.mirror_dir)
                    a = read_from_file(fn)
                    if a is None:
                        continue
                    a['filename'] = rel_fn
                    adn = os.path.dirname(rel_fn)
                    if rel_fn in ts:
                        a['tags'] = ts[rel_fn]
                    else:
                        a['tags'] = set()
                    if adn in rv:
                        rv[adn].stories.append(a)
                    else:
                        rv[adn] = AuthorInfo(os.path.abspath(
                            os.path.dirname(fn)))
                        rv[adn].stories.append(a)
        return rv

    def make_cache(self):
        """Calls read_entries(), stores the result (pickled) in the file index.db. This
        takes a long time on a large database.

        """
        with open(os.path.join(self.mirror_dir, "index.db"), 'wb') as fcache:
            pickle.dump(self.read_entries(), fcache, 3)

    def get_index(self, check=True):
        """Checks if the cache created by make_cache is up to date; if not, updates it.
        Either way, returns the index dictionary created by read_entries(). If
        check is false, return the cache unconditionally (used in the Web app).

        """
        cache_fn = os.path.join(self.mirror_dir, "index.db")
        ls = max(((i, os.stat(os.path.join(self.mirror_dir, i))) for i in
                  os.listdir(self.mirror_dir)), key=lambda x: x[1].st_mtime)
        if check and ls[0] != "index.db":
            a = self.read_entries()
            with open(cache_fn, 'wb') as fcache:
                pickle.dump(a, fcache, 3)
        else:
            with open(cache_fn, 'rb') as fcache:
                a = pickle.load(fcache)
        return a



# Site module for Archive Of Our Own.

import re
from bs4 import BeautifulSoup  # type: ignore
from bs4.element import Tag  # type: ignore
# from urllib.parse import urljoin
from datetime import datetime
import html2text  # type: ignore

from typing import Dict, Any, Tuple, List, Set, Optional

from ..util import urlopen_retry, rectify_strings, make_filename

def parse_int(s):
    return int(s.replace(',', ''))

class AO3(object):
    hostname = "archiveofourown.org"
    this_site = "ao3"
    story_url = "https://{hostname}/works/{number}"
    user_url = "https://{hostname}/users/{aid}/works"
    user_bookmarks_url = "https://{hostname}/users/{aid}/bookmarks"

    def get_user_url(self, md: Dict[str, Any]) -> str:
        return self.user_url.format(aid=md['authorid'], hostname=self.hostname)

    def get_story_url(self, md: Dict[str, Any]) -> str:
        return self.story_url.format(number=md['id'], hostname=self.hostname)

    def _parse_html_entry(self, i: Tag) -> Optional[Dict[str, Any]]:
        rv = {}
        o = i.p
        if o is not None and 'message' in o['class']:
            return None
        hh = i.find('h4', class_='heading')
        rv['title'] = hh.a.string
        if 'series' in hh.a['href']:
            return None
        o = re.search(r'\d+', hh.a['href'])
        assert o is not None
        rv['id'] = o.group(0)
        al = hh.find('a', rel='author')
        rv['author'] = al.string
        o = re.match(r'/users/(\w+)/', al['href'])
        assert o is not None
        rv['authorid'] = o.group(1)
        fandoms = [l.string for l in i.find('h5', class_='fandoms')
                   ('a', class_='tag')]
        rv['category'] = ' & '.join(fandoms)
        rv['updated'] = int(datetime.
                            strptime(i.find('div', class_='header').
                                     find('p', class_='datetime').string,
                                     "%d %b %Y").timestamp())
        rv['tags'] = set(fandoms)

        rt = i.find('ul', class_='required-tags')
        rating = rt.find('span', class_='rating').get_text().strip().lower()
        rv['tags'].add(f"rating: {rating}")
        rels = rt.find('span', class_='category')['title'].split(', ')
        rv['tags'].update([f"relationships: {l}" for l in rels])
        rv['complete'] = ('complete-yes' in rt.find('span', class_='iswip')
                          ['class'])

        tl = i.find('ul', class_='tags')
        for t in tl('li'):
            ts = t.find('a', class_='tag').string
            rv['tags'].add(ts)

        ht = html2text.HTML2Text()
        ht.ignore_links = True
        ht.ignore_tables = True
        ht.body_width = 0
        se = i.find('blockquote', class_='summary')
        se.name = 'div'
        rv['summary'] = ht.handle(str(se))
        se = i.find('dl', class_='stats')
        rv['language'] = se.find('dd', class_='language').string
        rv['words'] = parse_int(se.find('dd', class_='words').string)
        rv['chapters'] = parse_int(se.find('dd', class_='chapters').string.
                                   split('/')[0])
        try:
            rv['reviews'] = parse_int(se.find('dd', class_='comments').
                                      get_text().strip())
        except Exception:
            rv['reviews'] = 0
        return rectify_strings(rv)

    def download_list(self, number: str) -> Tuple[List[Dict[str, Any]],
                                                  List[Dict[str, Any]],
                                                  Dict[str, Any]]:
        """Despite the name, "number" is just the string username from the URL."""
        url = self.user_url.format(hostname=self.hostname, aid=number)
        data = urlopen_retry(url)
        assert data is not None
        soup = BeautifulSoup(data.read(), 'html5lib')

        def download_with_pages(url):
            rv = []
            page = urlopen_retry(url).read()
            soup = BeautifulSoup(page, 'html5lib')

            def make_page(n):
                return url + '?page=' + str(n)

            def get_page(u):
                if 'page=' not in u:
                    return 1
                else:
                    return int(re.search(r'page=(\d+)', u).group(1))

            pl = soup.find('ol', class_='pagination')
            if pl:
                max_page = max(get_page(i['href']) for i in pl('a'))
            else:
                max_page = 1

            def parse_page(soup):
                e = soup.find('div', id='main').find('ol', class_='index')
                for i in e('li', class_='blurb'):
                    try:
                        r = self._parse_html_entry(i)
                        if r is not None:
                            rv.append(r)
                    except Exception:
                        print(i)
                        raise

            parse_page(soup)
            for i in range(2, max_page + 1):
                u = make_page(i)
                s = BeautifulSoup(urlopen_retry(u).read(), 'html5lib')
                parse_page(s)

            return rv

        auth = download_with_pages(url)
        fav_url = self.user_bookmarks_url.format(hostname=self.hostname,
                                                 aid=number)
        fav = download_with_pages(fav_url)

        aname = (soup.find('div', id='main').find('h2', class_='heading').
                 string.split(' Works by ')[1].strip())
        author_dir = "{}-{}-{}".format(make_filename(aname), self.this_site,
                                       number)
        info = { 'author': aname, 'authorid': number, 'site': self.this_site,
                 'author_url': url, 'author_dir': author_dir }
        return auth, fav, info

    def get_tags_for(self, md: Dict[str, Any]) -> Set[str]:
        return md['tags']

    def download_metadata(self, number):
        url = self.story_url.format(hostname=self.hostname, number=number)



# Utilities common to the fanfic site scrapers

import time, urllib.request, urllib.error, re, hashlib, os, json
import urllib.parse, requests
try:
    import cloudscraper
except Exception:
    cloudscraper = None
from bs4 import NavigableString  # type: ignore
from typing import Dict, Optional, Any

def fold_string_indiscriminately(s: str, n: int = 80) -> str:
    """Folds a string (insert line-breaks where appropriate, to format
    on a display of no more than n columns) indiscriminately, meaning
    lose all existing whitespace formatting. This is the equivalent of
    doing an Emacs fill-paragraph on the string in question, though it
    doesn't break around double linefeeds like that function does."""
    l = s.split()
    rv = ""
    cl = 0
    for i in l:
        if cl + len(i) + 1 < n:
            rv += ' ' + i
            cl += len(i) + 1
        else:
            rv += ' \n' + i
            cl = len(i)
    return rv[1:]  # remove extraneous leading space

def fold_string_discriminately(s: str, n: int = 80) -> str:
    """Folds a string discriminately, that is, preserving existing
    hard line breaks in the original. This is the equivalent of
    passing the string to fold -s."""
    l = s.splitlines()
    rv = ""
    for i in l:
        if len(i) < n:
            rv += i + '\n'
        else:
            rv += fold_string_indiscriminately(i, n) + '\n'
    return rv

def make_filename(title: str) -> str:
    title = title.lower().replace(" ", "_")
    return re.sub("[^a-z0-9_.-]", "", title)

def url_hash(url: str) -> str:
    return hashlib.sha256(url.encode()).hexdigest()

class FakeRequest:
    def __init__(self, d: bytes) -> None:
        self.data = d

    def read(self) -> bytes:
        return self.data

class NetworkFetcher:
    def __init__(self, time_delay: float = 2.0):
        self.last_fetch: Dict[str, float] = {}
        self.delay = time_delay

    def do_fetch(self, url: str, timeout: int = 30,
                 use_cloudscraper: bool = False) -> bytes:
        headers = { "User-Agent":
                    # "Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:21.0) "
                    # "Gecko/20100101 Firefox/21.0" }
                    "Mozilla/5.0 (X11; Fedora; Linux x86_64; rv:76.0) "
                    "Gecko/20100101 Firefox/76.0" }
        host = urllib.parse.urlsplit(url).netloc
        if host not in self.last_fetch:
            self.last_fetch[host] = 0
        wait = time.time() - self.last_fetch[host]
        if wait < self.delay:
            time.sleep(self.delay - wait)
        self.last_fetch[host] = time.time()

        if use_cloudscraper:
            if cloudscraper is None:
                raise ValueError("cloudscraper not supported")
            fetcher = cloudscraper.create_scraper(
                browser={
                    'browser': 'firefox',
                    'platform': 'windows',
                    'mobile': False,
                    'desktop': True,
                })
            headers = {}
        else:
            fetcher = requests
        r = fetcher.get(url, headers=headers, timeout=timeout)
        r.raise_for_status()

        return r.content

def urlopen_retry(url: str, tries: int = 3, delay: float = 1.0,
                  timeout: int = 30,
                  cache_dir: str = '/home/tom/.ffmirror_cache',
                  fetcher: NetworkFetcher =
                  NetworkFetcher(),
                  use_cloudscraper: bool = False) -> Optional[FakeRequest]:
    """Open a URL, with retries on failure. Spoofs user agent to look like Firefox,
    since FFnet 403s the urllib UA."""
    fn = None
    if cache_dir is not None:
        uh = url_hash(url)
        fn = os.path.join(cache_dir, uh)
        if os.path.exists(fn) and os.stat(fn).st_mtime > time.time() - 43200:
            try:
                with open(fn) as inp:
                    r = json.load(inp)
                return FakeRequest(r['data'].encode())
            except Exception:
                pass
    for i in range(tries):
        try:
            # r = open_func(req, timeout=timeout)
            data = fetcher.do_fetch(url, timeout,
                                    use_cloudscraper=use_cloudscraper)
        except urllib.error.URLError as e:
            if i == tries - 1:
                raise e
            time.sleep(delay)
        else:
            if fn is not None:
                try:
                    o = { 'data': data.decode() }
                except UnicodeDecodeError:
                    print(url)
                    print(repr(data))
                    raise
                with open(fn, 'w') as out:
                    json.dump(o, out)
                return FakeRequest(o['data'].encode())
            return FakeRequest(data)
    return None

def rectify_strings(d: Dict[str, Any]) -> Dict[str, Any]:
    for i in d:
        if isinstance(d[i], NavigableString):
            d[i] = str(d[i])
    return d



# This is the site module for FF.net, and the model for other site modules. It
# handles only FF.net-specific tasks; the API provided includes
# download_metadata, compile_story, and download_list. The functions can be
# called on their own, but it's probably friendlier to use the runner in
# __main__ or the update functions in mirror. Also handles FictionPress, since
# they're identical.

import re, time
from bs4 import BeautifulSoup  # type: ignore
from bs4.element import Tag  # type: ignore
from datetime import datetime

from typing import Set, List, Tuple, Optional, Callable, TextIO

from ..util import (make_filename, urlopen_retry,
                    fold_string_indiscriminately)
from ..core import DownloadModule, StoryInfo, AuthorInfo, ChapterInfo

def cat_to_tagset(category: str) -> Set[str]:
    """Takes a category string, splits by crossover if necessary, returns a set of
    fandom tags for it. This will mangle category names with the substring
    ' & ' in them, but those are rare; I've seen only one ever.

    """
    rv = set()
    cl = category.split(' & ')
    for i in cl:
        # Tag name is category name, in lowercase, minus any commas
        rv.add(i.lower().replace(",", ""))
    return rv

# A site module needs to implement the following:
#  - get_user_url(self, md):
#    given a story metadata object, return a canonical link for the author
#  - get_story_url(self, md):
#    given a story metadata object, return a canonical link for the story
#  - download_metadata(self, sid):
#    given a story ID, download its metadata and TOC and return
#  - compile_story(self, md, toc, outfile):
#    given metadata and TOC, download a whole story and write it to outfile
#  - download_list(self, aid):
#    given author ID, download a list of stories as metadata (metadata entries
#    may be incomplete)
class FFNet(DownloadModule):
    hostname = "www.fanfiction.net"
    this_site = "ffnet"
    story_url = "https://{hostname}/s/{number}/{chapter}/"
    user_url = "https://{hostname}/u/{number}/"

    file_version = 3  # for metadata check

    url_re = re.compile(r"^https://" + hostname + "/")
    story_url_re = re.compile(r"https?://(?P<hostname>[^/]+)/s/(?P<sid>\d+)/?(?P<chl>\d+)?/?")  # noqa: E501
    user_url_re = re.compile(r"https?://(?P<hostname>[^/]+)/u/(?P<aid>\d+)/?")

    def get_user_url(self, md: AuthorInfo) -> str:
        return self.user_url.format(number=md.id,
                                    hostname=self.hostname)

    def get_story_url(self, md: StoryInfo) -> str:
        return self.story_url.format(number=md.id, chapter=1,
                                     hostname=self.hostname)

    # Functions related to downloading stories

    def _get_contents(self, soup: BeautifulSoup,
                      sid: str) -> List[ChapterInfo]:
        """Given a BeautifulSoup of a story page, extract the contents list and
        return list of chapter titles."""
        se = soup.find("select", id="chap_select")
        if se is None:
            # it's a oneshot, so no chapter list
            return [
                ChapterInfo(title='Chapter 1',
                            url=self.story_url.format(
                                hostname=self.hostname, number=sid, chapter=1
                            ))]
        rv: List[ChapterInfo] = []
        for n, oe in enumerate(se.find_all('option')):
            o = re.match(r"\d+. (.*)", oe.string)
            assert o is not None
            val = o.group(1)
            rv.append(
                ChapterInfo(title=val,
                            url=self.story_url.format(
                                hostname=self.hostname, number=sid, chapter=n
                            )))
        return rv

    # Apparently souping the storytext can mess some things up when FFn uses
    # bad HTML.

    # def get_storytext(soup):
    #     """Takes a soup of the page, extracts the story text HTML and returns
    #     it."""
    #     st = soup.find('div', id='storytext')
    #     d = str(st)
    #     d = re.sub(r"^<div[^>]*>", "", d)
    #     d = re.sub(r"</div>$", "", d)
    #     return d

    def _get_storytext(self, d: str) -> str:
        """Takes a page of HTML, extracts the storytext, returns it."""
        o = re.search("<div[^>]*id='storytext'[^>]*>", d)
        if o is None:
            raise Exception("Didn't find storytext")
        d = d[o.end():]
        o = re.search("</div>", d)
        if o is None:
            raise Exception("Didn't find storytext end")
        d = d[:o.start()]
        return d

    def _get_metadata(self, soup: BeautifulSoup) -> StoryInfo:
        """Given a BeautifulSoup of an FFnet page, extract a metadata entry. Somewhat
        abbreviated from what download_list returns, since it's parsing
        rendered HTML rather than JS, and some information is lost. This
        function need change often to handle FFnet's constant dumb alterations;
        I wish they'd provide an API.

        """
        e = soup.find("div", id="profile_top")
        title = e.find("b").string
        ae = e.find("a")
        author = ae.string
        o = re.match(r"/u/(\d+)/.*", ae['href'])
        assert o is not None
        authorid = o.group(1)
        sd = e.find("div", class_='xcontrast_txt')
        summary = sd.string
        md = sd.find_next_sibling("span", class_='xcontrast_txt')
        s = md.a.next_sibling
        o = re.match(r"[ -]+\w+[ -]+((?P<genre>[\w/]+) - )?((?P<chars>(?!Chapters).*?\S) +- +)?(Chapters: (?P<chaps>\d+)[ -]+)?Words: (?P<words>[\d,]+).*", s)  # noqa: E501
        assert o is not None
        characters = o.group('chars')
        genre = o.group('genre') or ''
        if characters is None:
            characters = ""
        try:
            chapters = int(o.group('chaps'))
        except TypeError:
            chapters = 1
        words = int(o.group('words').replace(",", ""))
        ae = md.a.find_next_sibling('a')
        if ae is None:
            reviews = 0
            ids = s
        else:
            reviews = int(ae.string.replace(",", ""))
            ids = ae.next_sibling
        ids = md.get_text()
        o = re.match(r".*id: (\d+).*", ids)
        assert o is not None
        sid = o.group(1)
        ud = md.find("span", attrs={"data-xutime": True})
        if ud.previous_sibling.endswith("Updated: "):
            updated = int(ud['data-xutime'])
            pd = ud.find_next_sibling("span", attrs={"data-xutime": True})
            published = int(pd['data-xutime'])
        elif ud.previous_sibling.endswith("Published: "):
            published = int(ud['data-xutime'])
            updated = published
        complete = 'Status: Complete' in md.get_text()
        e = soup.find("div", id='pre_story_links')
        try:
            category = e.a.find_next_sibling('a').string
        except AttributeError:
            category = 'crossover'
        author_url = self.user_url.format(hostname=self.hostname,
                                          number=authorid)
        story_url = self.story_url.format(hostname=self.hostname, number=sid,
                                          chapter=1)
        author_dir = "{}-{}-{}".format(make_filename(author), self.this_site,
                                       authorid)
        authinf = AuthorInfo(name=author, id=authorid, url=author_url,
                             site=self.this_site, dir=author_dir)
        storyinf = StoryInfo(
            title=title, summary=summary, category=category, id=sid,
            reviews=reviews, chapters=chapters, words=words,
            characters=characters, source='story', author=authinf, genre=genre,
            site=self.this_site, updated=datetime.fromtimestamp(updated),
            published=datetime.fromtimestamp(published),
            complete=complete, story_url=story_url,
            tags=cat_to_tagset(category)
        )
        return storyinf

    def _make_toc(self, contents: List[ChapterInfo]) -> str:
        """Makes an HTML string table of contents to be concatenated into outstr, given
        the return value of _get_contents (array of chapter names).

        """
        rs = "<h2>Contents</h2>\n<ol>\n"
        for x in range(len(contents)):
            n = x + 1
            anc = "#ch{}".format(n)
            rs += "<li><a href=\"{}\">{}</a></li>\n".format(anc,
                                                            contents[x].title)
        rs += "</ol>\n"
        return rs

    def download_metadata(self, number: str) -> Tuple[StoryInfo,
                                                      List[ChapterInfo]]:
        """This function takes a fic number and returns a pair: the first is a
        dictionary of the story's metadata, the second is sufficient
        information to download all its individual chapters.

        """
        url = self.story_url.format(hostname=self.hostname, number=number,
                                    chapter=1)
        r = urlopen_retry(url, use_cloudscraper=True)
        assert r is not None
        data = r.read()
        soup = BeautifulSoup(data, 'html5lib')
        md = self._get_metadata(soup)
        toc = self._get_contents(soup, number)
        return md, toc

    def compile_story(self, md: StoryInfo, toc: List[ChapterInfo],
                      outfile: TextIO, contents: bool = False,
                      callback: Optional[Callable[[int, str], None]] =
                      None) -> None:
        """Given the output of download_metadata, download all chapters of a story and
        write them to outfile. Extra keyword arguments are ignored in order to
        facilitate calls. callback is called as each chapter is downloaded with
        the chapter index and title; this should be a quick function to print
        progress output or similar, since its completion blocks continuing the
        download.

        """
        outfile.write("""<html>
<head>
<meta charset="UTF-8" />
<meta name="Author" content="{author}" />
<title>{title}</title>
<style type="text/css">
body {{ font-family: sans-serif }}
</style>
</head>
<!-- Fic ID: {id}
""".format(title=md.title, id=md.id, author=md.author.name))
        for k, v in sorted(md.to_dict().items()):
            outfile.write("{}: {}\n".format(k, v))
        outfile.write("-->\n<body>\n")
        outfile.write("<h1>{}</h1>\n".format(md.title))
        if contents:
            outfile.write(self._make_toc(toc))
        for n, t in enumerate(toc):
            time.sleep(2)
            x = n + 1
            url = self.story_url.format(hostname=self.hostname,
                                        number=md.id, chapter=x)
            if callback:  # For printing progress as it runs.
                callback(n, t.title)
            r = urlopen_retry(url, use_cloudscraper=True)
            assert r is not None
            data = r.read().decode()
            text = self._get_storytext(data)
            outfile.write("""<h2 id="ch{}" class="chapter">{}</h2>\n""".
                          format(x, t.title))
            text = fold_string_indiscriminately(text)
            outfile.write(text + "\n\n")
        outfile.write("</body>\n</html>\n")

    # Functions related to dealing with user listings

    def _parse_entry(self, elem: Tag) -> StoryInfo:
        """Given a BeautifulSoup element for a story listing, return a metadata object
        for the story.

        """
        title = elem['data-title'].replace("\\'", "'")
        category = elem['data-category'].replace("\\'", "'")
        sid = elem['data-storyid']
        published = int(elem['data-datesubmit'])
        updated = int(elem['data-dateupdate'])
        reviews = int(elem['data-ratingtimes'])
        chapters = int(elem['data-chapters'])
        words = int(elem['data-wordcount'])
        sd = elem.find('div')
        summary = sd.contents[0]
        sd = sd.find('div')
        o = re.match(r"(Crossover - )?(?P<category>.+?) - Rated: (?P<rating>.{1,2}) - (?P<language>.+?) - (?P<genre>.+?) - ", sd.contents[0])  # noqa: E501
        assert o is not None
        genre = o.group('genre')
        if 'Chapters' in genre:
            genre = ''
        cs = sd.contents[-1]
        if type(cs) == Tag:
            chars = ''
        else:
            o = re.match(r"\s*-\s+(.+?)(\s+-.*)?$", cs)
            assert o is not None
            chars = o.group(1)
        if chars == 'Complete':
            chars = ''
        complete = elem['data-statusid'] == '2'
        source = 'favorites' if 'favstories' in elem['class'] else 'authored'
        if source == 'favorites':
            al = elem.find('a', href=re.compile(r"^/u/.*"))
            author = al.string
            o = re.match(r"^/u/(\d+)/.*", al['href'])
            assert o is not None
            authorid = o.group(1)
            author_url = self.user_url.format(hostname=self.hostname,
                                              number=authorid)
            author_dir = "{}-{}-{}".format(make_filename(author),
                                           self.this_site, authorid)
        else:  # in this case, the caller populates those fields
            author = ''
            authorid = 0
            author_url = ''
            author_dir = ''
        story_url = self.story_url.format(hostname=self.hostname, number=sid,
                                          chapter=1)
        authinfo = AuthorInfo(name=author, id=authorid, url=author_url,
                              site=self.this_site, dir=author_dir)
        storyinfo = StoryInfo(
            title=title, category=category, id=sid,
            published=datetime.fromtimestamp(published),
            updated=datetime.fromtimestamp(updated),
            reviews=reviews, chapters=chapters, words=words, summary=summary,
            characters=chars, complete=complete, source=source,
            author=authinfo, genre=genre, site=self.this_site,
            story_url=story_url, tags=cat_to_tagset(category)
        )
        return storyinfo

    def download_list(self, number: str) -> Tuple[List[StoryInfo],
                                                  List[StoryInfo],
                                                  AuthorInfo]:
        """Given a user ID, download lists of the stories they've written and favorited
        and return them. The lists are returned as a tuple of (authored,
        faved). Each entry is a dictionary containing metadata.

        """
        try:
            url = self.user_url.format(hostname=self.hostname, number=number)
            r = urlopen_retry(url, use_cloudscraper=True)
            assert r is not None
            page = r.read().decode()
            soup = BeautifulSoup(page, 'html5lib')
            author = (soup.find('div', id='content_wrapper_inner').span.string.
                      strip())
            auth: List[StoryInfo] = []
            fav: List[StoryInfo] = []
            author_dir = "{}-{}-{}".format(make_filename(author),
                                           self.this_site,
                                           number)
            for i in soup.find_all('div', class_='z-list'):
                a = self._parse_entry(i)
                if a.source == 'favorites':
                    fav.append(a)
                elif a.source == 'authored':
                    a.author.name = author
                    a.author.id = number
                    a.author.dir = author_dir
                    a.author.url = url
                    auth.append(a)
            info = AuthorInfo(name=author, id=number, site=self.this_site,
                              url=url, dir=author_dir)
        except Exception:
            print(url)
            raise
        return auth, fav, info

    def compare_mds(self, r: StoryInfo, cr: StoryInfo) -> bool:
        """Check two metadata entries to see if they 1. refer to the same story, 2. at
        the same length. Used to check for updates."""
        if (r.words != cr.words or r.chapters != cr.chapters or
            r.updated != cr.updated):  # noqa: E129
            return False
        return True

class FictionPress(FFNet):
    hostname = "www.fictionpress.com"
    this_site = "fictionpress"
    url_re = re.compile(r"^https://" + hostname + "/")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ffmirror', 'ffmirror.handlers']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.42,<2.0.0',
 'attrs>=19.3,<20.0',
 'beautifulsoup4>=4.7,<5.0',
 'click>=8.1.3,<9.0.0',
 'html2text>=2019.9,<2020.0',
 'html5lib>=1.0,<2.0',
 'python-dateutil>=2.7,<3.0',
 'requests>=2.22,<3.0']

entry_points = \
{'console_scripts': ['ffdb = ffmirror.cli:run_db_op',
                     'ffdl = ffmirror.cli:run_dl']}

setup_kwargs = {
    'name': 'ffmirror',
    'version': '0.4.0',
    'description': 'Local mirror for Internet fiction sites',
    'long_description': "ffmirror is a program to create and maintain a local mirror of stories on\nfiction-publishing sites. It has functionality for downloading stories similar\nto FanFicFare or other projects, but the differentiating factor is support for\ncreating a local database that maintains downloaded metadata and can\nautomatically update authors that are followed.\n\nffmirror has two currently maintained script entry points:\n\n - ffdl is a simple one-story file downloader. It takes a single URL and writes\n   an HTML file story.\n - ffdb is the manager for local fanfic site mirrors. These maintain a local\n   copy of metadata for a set of followed users alongside copies of all their\n   stories.\n\nThe remaining entry points, ffadd, ffup, ffcache are designed for an older\nformat of mirror, and are now deprecated.\n\nffmirror can be installed via PyPI: ``pip install ffmirror``\n\nTo create a mirror, enter an empty directory and issue:\n\n.. code:: bash\n\n  $ ffdb init\n\nThis initializes the SQLite database that tracks metadata. You can now add\nauthors by issuing:\n\n.. code:: bash\n\n  $ ffdb add $AUTHOR_URL\n\nAdding an author will immediately download all their stories into the mirror.\nStories are stored under top-level directories per author.\n\nUpdating the mirror will recheck all authors that have been added and download\nany new or updated stories. To update the mirror, issue:\n\n.. code:: bash\n\n  $ ffdb update\n\nYou can update only one author by issuing:\n\n.. code:: bash\n\n  $ ffdb update $AUTHOR_DIR\n\nwhere AUTHOR_DIR is the directory with that author's stories.\n",
    'author': 'alethiophile',
    'author_email': 'tomdicksonhunt@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alethiophile/ffmirror',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
